#!BPY

"""
Name: 'SMD Animation Importer'
Blender: 2.41
Group: 'Import'
Tip: 'import a HL2 Animation (.smd) File'
"""

__author__="puh"

__version__="1.00"

__bpydoc__="""import_smd_animation.py

HL Animation Importer by puh 

based on:

OGRE3D Mesh Importer by Daniel (D-Man) Handke
[OGRE3D_Mesh_importer.py]

and

Half-Life 2 Static SMD Importer Blender Script by Jon Moen Drange
[import_static_smd.py]



imports HL Animation .smd files

It's an importer for HL Animation Files

Press Alt-P to start the script.

"""
############################ CODE STARTS HERE ############################

import Blender, meshtools, bpy, math, time
from Blender import *
from Blender.Mathutils import *


########################################################
##### Exceptionclasses #################################

class ParsingSMDFileError (Exception):
   def __init__(self,value):
      Exception.__init__(self,value)


########################################################
##### Base class for animation frames ##################
class Keyframe:
   def __init__(self,number,posX,posY,posZ,rotX,rotY,rotZ):
      self.number = number + 1
      self.posX     = posX
      self.posY     = posY
      self.posZ     = posZ
      self.rotX     = rotX
      self.rotY     = rotY
      self.rotZ     = rotZ
      self.currentVector = Vector([posX,posY,posZ,1])
      self.currentRotMat = (RotationMatrix(rotX,4,'x')
                          * RotationMatrix(rotY,4,'y')
                          * RotationMatrix(rotZ,4,'z'))
      #TODO add more variables


########################################################
##### Small classes for various purposes ###############

class Bone:
   def __init__(self,number,name,parentnumber,bones):
      self.childs   = []
      self.index    = number
      self.name     = name
      self.indexPar = parentnumber
      self.parent   = None
      if (self.indexPar > -1):
         self.parent   = bones[self.indexPar]
         #self.bone.parent = bones[self.indexPar].bone
      self.matrix   = None
      self.frames = []

   def __repr__(self):
      return '['+str(self.index)+' "'+self.name+'" Par: '+str(self.indexPar)+']'
   def __str__(self):
      return self.__repr__()


########################################################
##### Base class for all importers #####################
class Importer:
   def __init__(self, filename):
      # making file ready for reading
      self.filename = filename
      self.file = open(filename, "rb")
      self._lines = self.file.readlines()
      self._lineno = 0

      # useful vars / pointers
      self.armObj  = None
      self.armature= None
      self.bones   = []
      self.scene   = Blender.Scene.GetCurrent()
      self.frames   = []
      self.noFrame = 0
      self.noBones = 0
      self.time0   = 0
      self.scale   = 1.0 / 40.0
      self.scaleMat = Matrix([self.scale,0,0,0],
                             [0,self.scale,0,0],
                             [0,0,self.scale,0],
                             [0,0,0,         1])

      #TODO add more variables
      
      

   def printDebugInfo(self):
      print '\nDEBUGINFO for SMD Importer (fixme!)\n'


   def getLine(self):
      if (self._lineno < len(self._lines)):
         tmp = self._lines[self._lineno].split("\r\n",1)[0]
         self._lineno += 1
         return tmp
      else:
         return ''


########################################################
##### class for importing armature and animation ###
class StaticMeshImporter (Importer):
   def __init__(self, filename):
      Importer.__init__(self, filename)


   def importSMD(self):
      if ( self.getLine() != 'version 1' ):
         raise ParsingSMDFileError, 'First line should be "version 1"'
      self.parseNodes()
      self.parseArmature()
      self.createActions()


   ### parsing the nodes data from SMD file
   def parseNodes(self):
      print 'Parsing Nodes ...'
      line = self.getLine()
      if (line != 'nodes'):
         raise ParsingSMDFileError,('Expecting section with nodes data'
                              +' at line '+str(self._lineno))
      
      line = self.getLine()
      while ( line != 'end'):
         line = line.split()
         boneindex   = int(line[0])
         parentindex = int(line[ len(line)-1 ])
         
         tmp = '' # finding the name of the bone
         for i in range (1,len(line)-1):
            if i == 1: tmp += str(line[i])
            else: tmp += ' ' + str(line[i])
         tmp = tmp[1:]; tmp = tmp[:-1]
         
         if (boneindex != self.noBones):
            raise ParsingSMDFileError,('Inconsistent bone-indices in SMD file'
                                 +' at line '+str(self._lineno))
         self.bones.append(Bone(self.noBones,tmp,
                                parentindex,self.bones))
         self.noBones += 1
         
         line = self.getLine()
      # end while

      if DEBUG: #printing bones
         print 'BONESDEBUG'
         for i in range (0,len(self.bones)):
            print str(i) + ' ->  ' + str(self.bones[i])

      print 'Parsing Nodes: OK'


   ### parsing the skeleton data from SMD file
   def parseArmature(self):
      print 'Parsing Animation Frames ...'
      line = self.getLine()
      if (line != 'skeleton'):
         raise ParsingSMDFileError, ('Expecting section with skeleton data'
                                 +' at line: '+str(self._lineno))

      mode = 0
      line = self.getLine()
      while (line != 'end'):

         if ( (mode % (self.noBones+1)) == 0):
            #TODO make sure ONLY "time 0"
            if DEBUG: print line
            line = line.split()
            frame = int(line[1])
            print frame            
            mode = 1; line = self.getLine()
            continue # to next iteration

         line = line.split()
         if (len(line) != 7):
            raise ParsingSMDFileError, ('Expected 7 numbers (1 int +'
                                 +' 6 floats) at line '+str(self._lineno))

         # reading from current line
         boneNo = int(line[0])
         posX = float(line[1])
         posY = float(line[2])
         posZ = float(line[3])
         rotX = math.degrees( float(line[4]) )
         rotY = math.degrees( float(line[5]) )
         rotZ = math.degrees( float(line[6]) )
         # creating some matrices for current bone
         current = self.bones[boneNo]
         current.frames.append(Keyframe(frame,posX,posY,posZ,rotX,rotY,rotZ))
         mode += 1
         line = self.getLine() # end while

      print 'Parsing Animation Frames: OK'
      return
      # took out finalsetup armature

   def createActions(self):
       print 'Creating Action'

       # getting existing armature
       armobj = Object.Get('imported_armature')

       armdata = armobj.getData()
       pose = armobj.getPose()

       newaction = Armature.NLA.NewAction('imported_action')
       newaction.setActive(armobj)
       for bone in self.bones:
            rest_bone= armdata.bones[bone.name]
            pbone = pose.bones[bone.name]
            for frm in bone.frames:
                print "Bone: ",bone.name," Frame N: ",frm.number
                # calculating were the bone should be
                if (bone.indexPar == -1):
                   print "Bone has no parent"
                   # calculating when has no parent
                   matrix1 = frm.currentRotMat * TranslationMatrix(frm.currentVector)
                   headVec =(Vector([0,0,0,1])*matrix1).resize3D()
                   tailVec =(Vector([0.05,0,0,1])*matrix1).resize3D()
                else:
                   print "Bone has parent"
                   # calculating when has parent
                   matrix1 = frm.currentRotMat * TranslationMatrix(frm.currentVector)
                   parent  = bone.parent
                   matrix2 = parent.frames[frm.number-1].currentRotMat * TranslationMatrix(parent.frames[frm.number-1].currentVector)
                   matrix1 = matrix1 * matrix2
                   headVec = (Vector([0,0,0,1])*matrix1).resize3D()
                   tailVec = (Vector([0.15,0,0,1])*matrix1).resize3D()

                dirVec = (tailVec - headVec).normalize()

                #pbone.quat = matrix1.toQuat()
                pbone.quat = dirVec.toTrackQuat("-y", "z")
                pbone.insertKey(armobj,frm.number,[Object.Pose.ROT])
                #pbone.loc = frm.currentVector
                #pbone.insertKey(armobj,frm.number,[Object.Pose.LOC])

   ### cleaning up
   def cleanup(self):
      print ' ' # todo: maybe fix me
      


########################################################
##### Variouse Functions ###############################

def userAlert (msg):
   Blender.Draw.PupMenu ( 'An Error Occured%t|'+msg )


### Control-function for the import script #############
def myImport (filename):
   # Printing info in console
   print '\n\nHalf-Life 2 SMD Animation Importer ('+__version__+')'
   print '**********************************************'

   importer = None
   error = False

   try:
      importer = StaticMeshImporter(filename)
      # starting parsing file in this call
      importer.time0 = time.time() # for taking time
      importer.importSMD()
      
      
   except ParsingSMDFileError, e:
      print '\nERROR (Parsing file):\n' + str(e)
      userAlert ('Import failed. Check console!')
      error = True
   except IOError, e:
      print '\nERROR (IOError):\n' + str(e)
      userAlert ('Import failed. Check console!')
      error = True
   except Exception, e:
      print '\nERROR (Unknown):\n' + str(e)
      userAlert ('Import failed. Check console!')
      error = True

   # Finishing
   if (error == True):
      #if importer != None:
         #print 'Trying to clean up unfinished imported objects...'
         #importer.cleanup()
      print '\nScript terminated. There might be unfinished'
      print 'imported objects in your current project.\n'
   else:
      print '\nImporting Completed!'
      print 'Total time taken: '+str(int(time.time()-importer.time0))+' seconds'
   print '\n**********************************************\n'

# Global variables
DEBUG = False  # whether to print debug information to console or not

##### Script starts with a fileselector
Blender.Window.FileSelector (myImport, 'Import SMD Animation ...')